/* SPDX-License-Identifier: GPL-2.0 */
#ifndef PERF_MEMSWAP_H_
#define PERF_MEMSWAP_H_

void mem_bswap_64(void *src, int byte_size);
void mem_bswap_32(void *src, int byte_size);

#endif /* PERF_MEMSWAP_H_ */
