/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _LIBLOCKDEP_LINUX_HARDIRQ_H_
#define _LIBLOCKDEP_LINUX_HARDIRQ_H_

#define SOFTIRQ_BITS	0UL
#define HARDIRQ_BITS	0UL
#define SOFTIRQ_SHIFT	0UL
#define HARDIRQ_SHIFT	0UL
#define hardirq_count()	0UL
#define softirq_count()	0UL

#endif
