/* SPDX-License-Identifier: GPL-2.0 */

#ifndef PERF_DWARF2_H
#define PERF_DWARF2_H

/* dwarf2.h ... dummy header file for including arch/x86/lib/mem{cpy,set}_64.S */

#define CFI_STARTPROC
#define CFI_ENDPROC
#define CFI_REMEMBER_STATE
#define CFI_RESTORE_STATE

#endif	/* PERF_DWARF2_H */

