// SPDX-License-Identifier: GPL-2.0

#include <babeltrace/ctf-writer/writer.h>
#include <babeltrace/ctf-ir/stream-class.h>

int main(void)
{
	bt_ctf_stream_class_get_packet_context_type((void *) 0);
	return 0;
}
